@@ -63,7 +63,7 @@
   #:use-module ((nyacc lex) #:select (cnumstr->scm))
   #:use-module ((nyacc util) #:select (ugly-print))
   #:re-export (*nyacc-version*)
-  #:version (2 01 2))
+  #:version (2 01 3))
 
 ;; maybe change to a record-type
 (define *options* (make-parameter '()))
@@ -411,8 +411,8 @@
        (call-with-values
            (lambda () (let ((return (mdecl->udecl (cons "~ret" tail))))
                         (function*-wraps return params)))
-         (lambda (wrapper unwrapper)
-           `(cpointer (cfunction ,wrapper ,unwrapper)))))
+         (lambda (pc->pr pr->pc)
+           `(cpointer (cfunction ,pc->pr ,pr->pc)))))
       (`((pointer-to) (pointer-to) (function-returning . ,rest) . ,rest)
        `(cpointer (cbase 'void*)))
       (`((pointer-to) (struct-ref (ident ,name)))
@@ -594,16 +594,35 @@
 
 (define (unwrap-mdecl mdecl)
   (let ((defined (*defined*))
-        (mname (string->symbol (md-label mdecl))) (mtail (md-tail mdecl)))
+        (mname (string->symbol (md-label mdecl)))
+        (mtail (md-tail mdecl)))
     (match (car mtail)
       (`(fixed-type ,name) `(unwrap-number ,mname))
       (`(float-type ,name) `(unwrap-number ,mname))
-      (`(void) #f)
       (`(typename ,name)
        (cond
 	((dmem? name def-defined) `(unwrap-number ,mname))
 	((dmem? name defined) (defined-type-unwrapper name mname))
 	(else #f)))
+      (`(pointer-to)
+       (match mtail
+         (`((pointer-to) (typename ,name))
+          (cond
+           ((dmem? name def-defined) `(unwrap-pointer ,mname))
+           ((dmem? (w/* name) defined)
+            `(unwrap-pointer ,mname ,(strings->symbol name "*")))
+           ((dmem? name defined)
+            `(unwrap-pointer ,mname (cpointer ,(string->symbol name))))
+           (else `(unwrap-pointer ,mname))))
+         (`((pointer-to) (function-returning (param-list . ,params)) . ,rest)
+          (let ((return (mdecl->udecl (cons "~ret" rest))))
+            (call-with-values
+                (lambda () (function*-wraps return params))
+              (lambda (pr pc)
+                (let ((pc '(lambda (p) 'unused))) ; only proc->ptr is used
+                  `(unwrap-pointer ,mname (cpointer (cfunction ,pr ,pc))))))))
+         (`((pointer-to) . ,_)
+          `(unwrap-pointer ,mname))))
       (`(enum-def (ident ,name) ,_)
        (cond
 	((dmem? (w/enum name) defined)
@@ -617,10 +636,10 @@
 	(else `(unwrap~enum ,mname))))
       (`(struct-ref (ident ,name)) `(cdata&-ref ,mname))
       (`(union-ref (ident ,name)) `(cdata&-ref ,mname))
-      (`(pointer-to) `(unwrap-pointer ,mname))
       (`(array-of ,size) `(unwrap-array ,mname))
       (`(array-of) `(unwrap-array ,mname))
       ;; not expected
+      (`(void) #f)
       (`(struct-def . ,_) `(cdata&-ref ,mname))
       (`(union-def . ,_) `(cdata&-ref ,mname))
       (otherwise
@@ -733,8 +752,8 @@
 ;; Based on return udecl @var{return} and udecl params @var{params},
 ;; generate two procedures:
 ;; @enumerate
-;; @item generate procedure from pointer
 ;; @item generate pointer from procedure
+;; @item generate procedure from pointer
 ;; @end enumerate
 ;; @end deffn
 (define (function*-wraps return params)
@@ -751,13 +770,11 @@
          (va-call `(apply ~proc ,names (map cdr ~rest))))
     (values
      ;; procedure->pointer
-     `(lambda (~proc)
-        (ffi:procedure->pointer
-         ,decl-ret
-         (lambda ,names
-           (let ,wrap-par
-             ,(if cbak-ret `((lambda (~ret) ,cbak-ret) ,call) call)))
-         (list ,@decl-par)))
+     (let* ((call (if cbak-ret `((lambda (~ret) ,cbak-ret) ,call) call))
+            (wrap (if (pair? wrap-par) `(let ,wrap-par ,call) call)))
+       `(lambda (~proc)
+          (ffi:procedure->pointer
+           ,decl-ret (lambda ,names ,wrap) (list ,@decl-par))))
      (if varargs?
          ;; pointer->procedure (varargs)
          `(lambda (~fptr)
@@ -885,22 +902,22 @@
           (`((function-returning (param-list . ,params)) . ,rest)
            (let ((return (mdecl->udecl (cons "~ret" rest))))
              (call-with-values (lambda () (function*-wraps return params))
-               (lambda (ptr->proc proc->ptr)
+               (lambda (proc->ptr ptr->proc)
                  (values
                   (dcons name (w/* name) defined)
                   (xcons* seed
-                    (deftype type `(cfunction ,ptr->proc ,proc->ptr))
+                    (deftype type `(cfunction ,proc->ptr ,ptr->proc))
                     (deftype type* `(cpointer ,type))))))))
 
           (`((pointer-to) (function-returning (param-list . ,params)) . ,rest)
            (let ((return (mdecl->udecl (cons "~ret" rest))))
              (call-with-values (lambda () (function*-wraps return params))
                (let ((*type (strings->symbol "*" name)))
-                 (lambda (pr->pc pc->pr)
+                 (lambda (pc->pr pr->pc)
                    (values
                     (dcons name defined)
                     (xcons* seed
-                      `(define ,*type (cfunction ,pr->pc ,pc->pr))
+                      `(define ,*type (cfunction ,pc->pr ,pr->pc))
                       (deftype type `(cpointer ,*type)))))))))
 
           (`((pointer-to) . ,rest)
@@ -971,8 +988,7 @@
                        (deftype type* `(cpointer (delay ,type))))))
                   (else ;; not defined
                    (xcons* seed
-                     `(define ,type 'void)
-                     (deftype type* `(cpointer ,type))))))))
+                     (deftype type* `(cpointer 'void))))))))
 
              ((union-ref (ident ,agname))
               (let ((agname (rename agname 'type)))
@@ -991,8 +1007,7 @@
                        (deftype type* `(cpointer (delay ,type))))))
                   (else ;; not defined
                    (xcons* seed
-                     `(define-public ,type 'void)
-                     (deftype type* `(cpointer ,type))))))))
+                     (deftype type* `(cpointer 'void))))))))
 
              (((fixed-type float-type) ,basename)
               (values defined (cons (deftype type (mtail->ctype mtail)) seed)))
